/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Gopher Protocol Handler code.
 *
 * The Initial Developer of the Original Code is Cameron Kaiser
 * <ckaiser@floodgap.com>.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 * Based partially on code by Bradley Baetz <bbaetz@student.usyd.edu.au>
 * and Darin Fisher <darin@meer.net>.
 *
 * Contributors:
 *   Cameron Kaiser <ckaiser@floodgap.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//const GOPHER_PROTOCOL = "bletch";
const GOPHER_PROTOCOL = "gopher";
// netwerk/streamconv/converters/nsIndexedToHTML.cpp requires that the
// protocol name be gopher in order for the "magic" style changes.

const GOPHER_HANDLER_CONTRACTID = "@mozilla.org/network/protocol;1?name="+
                                   GOPHER_PROTOCOL;
const GOPHER_HANDLER_CID = Components.ID(
                           "{a892e4e0-4dc8-11dd-ae16-0800200c9a66}");

const GOPHER_MSG_BUNDLE_URL = "chrome://necko/locale/necko.properties";
// in mozilla/netwerk/locales/[locale]/necko.properties

const Ci = Components.interfaces;
const Cc = Components.classes;
const Cr = Components.results;
const Cu = Components.utils;

Cu.import("resource://gre/modules/XPCOMUtils.jsm");

const nsISupports = Ci.nsISupports;
const nsIProtocolHandler = Ci.nsIProtocolHandler;
const nsIProxiedProtocolHandler = Ci.nsIProxiedProtocolHandler;

const nsIRequest = Ci.nsIRequest;
const nsIChannel = Ci.nsIChannel;
const nsIProxiedChannel = Ci.nsIProxiedChannel;
const nsIProgressEventSink = Ci.nsIProgressEventSink;

const nsISocketTransport = Ci.nsISocketTransport;
const nsIStreamConverter = Ci.nsIStreamConverter;
const nsIStreamListener = Ci.nsIStreamListener;
const nsIObserver = Ci.nsIObserver;
const nsIRequestObserver = Ci.nsIRequestObserver;

const GOPHER_MENU_MIME_TYPE = "text/x-gopher-menu";
// see UMN Gopher+ for rationale for this MIME type's use

// whitelist for valid ports to speak on
const ValidGopherPorts =
                   [ 13, 43, 70, 71, 72, 79, 80, 105, 1070, 2347, 3000, 3099,
                     4323, 7055, 7070, 7071, 7072, 7077, 7080, 7777, 27070 ];
// 80 is included for hybrid servers like GN and pygopherd that can speak
// http or gopher on a single port. 2347 is included for veronica; 105 for
// cso/ph (itemtype 2); 13 for daytime (used on some systems); 43 for whois;
// 79 for finger. the remainder were sampled from the Veronica-2 list of valid
// servers still extant.

// global function for logging to the error console (commented out for
// future debuggerers)
function GopherLog(msg, error) {
/*
  return Cr.NS_OK;
  var consoleService = Cc["@mozilla.org/consoleservice;1"].
                       getService(Ci.nsIConsoleService);
  msg = "GopherFF says: "+msg;
  if (error) {
    consoleService.logStringError(msg);
  } else {
    consoleService.logStringMessage(msg);
  }
*/
}

/* The object that presents the protocol and filters initial requests.
   This is the only object directly offered through the factory (see below). */

function GopherProtocol() {
}

GopherProtocol.prototype = {
  // our stuff

  // l10n/i18n support -- load our string bundles
  _bundle : Cc["@mozilla.org/intl/stringbundle;1"].
            getService(Ci.nsIStringBundleService).
            createBundle(GOPHER_MSG_BUNDLE_URL),
  _getL10NString : function GP_getL10NString(msg, args) {
    if (args) 
      return this._bundle.formatStringFromName(msg, args, args.length);
    return this._bundle.GetStringFromName(msg);
  },

  classDescription : "Gopher protocol handler",
  contractID : GOPHER_HANDLER_CONTRACTID,
  classID : GOPHER_HANDLER_CID,

  // nsISupports
  QueryInterface : XPCOMUtils.generateQI([nsISupports,
                                         nsIProtocolHandler,
                                         nsIProxiedProtocolHandler]),

  // nsIProtocolHandler
  scheme: GOPHER_PROTOCOL,
  defaultPort: 70,
  protocolFlags: 0
      | nsIProtocolHandler.ALLOWS_PROXY
      | nsIProtocolHandler.ALLOWS_HTTP_PROXY
      | nsIProtocolHandler.URI_LOADABLE_BY_ANYONE
    ,
  
  allowPort : function GP_allowPort(port, scheme) {
    return (ValidGopherPorts.indexOf(port) != -1);
  },

  newURI : function GP_newURI(spec, charset, baseURI) {
    var uri = Cc["@mozilla.org/network/standard-url;1"].
              createInstance(Ci.nsIURI);
    uri.spec = spec;
    return uri;
  },

  newChannel : function GP_newChannel(input_uri) {
    GopherLog("new request for "+input_uri.asciiSpec);
    return this.newProxiedChannel(input_uri, null);
  },

  // nsIProxiedProtocolHandler 
  newProxiedChannel : function GP_newProxiedChannel(input_uri, proxyinfo) {
    var ioService = Cc["@mozilla.org/network/io-service;1"].
                    getService(Ci.nsIIOService);
    var prompter = Cc["@mozilla.org/embedcomp/prompt-service;1"].
                   getService(Ci.nsIPromptService);

    GopherLog("new proxied request for "+input_uri.asciiSpec);
    if (proxyinfo)
      GopherLog("proxy is: "+proxyinfo);

    // try to handle trivial cases immediately without making a channel ...

    // handle hURLs directly here (and reject things like javascript:, data:)
    // if we are passed a gopher URL with a hURL embedded in it
    if (input_uri.path.match(/^\/?h\/?URL:.+/)) {
      var newuri = input_uri.path.replace(/^\/?h\/?URL:\/?/, "");
      GopherLog("URL REDIRECT: "+newuri);

      // reject unsafe destination schemes
      if (newuri.match(/^javascript:/) || newuri.match(/^data:/)) {
        prompter.alert(null,
                       this._getL10NString('GopherhURLIllegalTitle', [newuri]),
                       this._getL10NString('GopherhURLIllegalText'));
        throw Cr.NS_ERROR_ABORT;
        return null;
      }

      var rv = prompter.confirm(null,
                                this._getL10NString('GopherhURLWarningTitle'),
                                this._getL10NString('GopherhURLWarningText',
                                                    [newuri]));
      if (rv) {
        // is there a better way to do this?

        var wm = Cc["@mozilla.org/appshell/window-mediator;1"].
                 getService(Ci.nsIWindowMediator);
        wm.getMostRecentWindow('navigator:browser').getBrowser().
           webNavigation.loadURI(newuri, 0, null, null, null);
        // this is a bit cheap, but we don't want
        // FF freaking out that we didn't give it a channel.
        throw Cr.NS_ERROR_ABORT;

        return null; // above will do redirect
      } else {
        throw Cr.NS_ERROR_ABORT;
        return null;
      }
    }

    // handle itemtype 7 at this stage and turn into a channel
    // for itemtype 1 instead, except if this particular URL has
    // arguments (but wouldn't you want to do that as itype 1?).
    if (input_uri.path.substr(0,2) == "/7" &&
        input_uri.path.indexOf("?") == -1) {
      var chequestub = { value : false };
      var query = { value : '' };

      // we will accept "blank" responses -- could be valid
      var rv = prompter.prompt(null,
                               this._getL10NString('GopherPromptTitle'),
                               this._getL10NString('GopherPromptText'),
                               query, null, chequestub);
      if (!rv) {
        throw Cr.NS_ERROR_ABORT;
        return null;
      }
      // stuff query into channel query args rather than
      // kludging it into a URL
      var ob = new GopherChannel(input_uri, proxyinfo);
      ob.queryargs = query.value;
      return ob;
    }

    // similarly handle itemtype 2
    // if it's /2....bletchfoo then pass that on to the CSO server
    // (it's a fully qualified query that requires no parsing)
    if (input_uri.path == "/2" || input_uri.path == "/2/") {
      var chequestub = { value : false };
      var query = { value : '' };

      // unlike itype 7 we MUST enter a query for this
      var rv = prompter.prompt(null,
                               this._getL10NString('GopherCSOPromptTitle'),
                               this._getL10NString('GopherCSOPromptText'),
                               query, null, chequestub);
      if (rv && !query.value.length) // blank query
        prompter.alert(null,
                       this._getL10NString('GopherCSOErrorTitle'),
                       this._getL10NString('GopherCSOErrorText'));
      if (!rv || !query.value.length) {
        throw Cr.NS_ERROR_ABORT;
        return null;
      }
      // stuff query into channel csoargs
      var ob = new GopherChannel(input_uri, proxyinfo);
      ob.csoargs = query.value;
      return ob;
    }
      
    // otherwise
    // immediately reject "pseudo" item types we don't handle
    var wontshow = ['/i', '/3', '/8', '/T'];
    if (input_uri.path && input_uri.path.length > 1 &&
                          wontshow.indexOf(input_uri.path.substr(0,2)) > -1) {
      prompter.alert(null,
                     this._getL10NString('GopherBadItemTitle'),
                     this._getL10NString('GopherBadItemText'));
      throw Cr.NS_ERROR_ABORT;
      return null;
    }

    // reject non-whitelisted port numbers
    if (input_uri.port && input_uri.port >= 0 &&
                          ValidGopherPorts.indexOf(input_uri.port)== -1) {
      GopherLog("illegal port: "+input_uri.port);
      throw Cr.NS_ERROR_PORT_ACCESS_NOT_ALLOWED;
      return null;
    }

    // else it's a legit gopher request
    // make our channel and gopher it
    return new GopherChannel(input_uri, proxyinfo);
  }
};

/* The channel object that actually does the protocol negotiation.
   This is not separately exposed since GopherProtocol is probably the only
   object that will ever call it. */

function GopherChannel(input_uri, proxyinfo) {
  this.init(input_uri, proxyinfo);
}

GopherChannel.prototype = {

  // my stuff
  _itemtype : null,
  _selector : null,
  _host : null,
  _port : null,
  _transport : null,
  _proxyinfo : null,
  _pump : null,
  _listener : null,
  _context : null,
  _progsink : null,

  // my internal functions
  _queryCallback : function GC_queryCallback(one, two, three) {
    var progsink = null;

    if (three)
      return three;
    if (one.notificationCallbacks) {
      progsink = one.notificationCallbacks.getInterface(nsIProgressEventSink);
      if (!progsink && two) {
        // try that instead
        var cbs = two.notificationCallbacks;
        if (cbs)
          progsink = cbs.getInterface(nsIProgressEventSink);
      }
    }
    return progsink;
  },

  // request arguments taken from user requesters
  csoargs : '', // arguments from query requester
  queryargs : '', // arguments from CSO query requester
  transreq : null, // actual transmitted request (see sendRequest)

  // nsISupports
  QueryInterface: XPCOMUtils.generateQI([nsISupports,
                                         nsIChannel,
                                         nsIProxiedChannel,
                                         nsIRequest]),

  // nsIRequest
  loadFlags : 0,
  name : null,
  isPending : false,
  _loadGroup : null,

  get status() {
    if (this._pump)
      return (this._pump.status);
    return this._status;
  },
  set status(aStatus) { this._status = aStatus; },
  cancel : function GC_cancel(aStatus) {
    if (this._pump)
      return (this._pump.cancel(aStatus));
    return Cr.NS_OK;
  },
  suspend : function GC_suspend() {
    if (this._pump)
      return (this._pump.suspend());
    return Cr.NS_OK;
  },
  resume : function GC_resume() {
    if (this._pump)
      return (this._pump.resume());
    return Cr.NS_OK;
  },
  get loadGroup() { return this._loadGroup; },
  set loadGroup(aLoadGroup) {
    this._loadGroup = aLoadGroup;
    this._progsink = null;
  },

  // nsIChannel
  loadAttributes : null,
  contentCharset : null,
  contentLength : -1,
  contentType : null,
  _notificationCallbacks : null ,
  originalURI : null,
  owner : null,
  URI : null,
  init : function GC_init(input_uri, proxyinfo) {
    // constructor
    this.URI = input_uri;
    this.originalURI = input_uri;
    this.name = input_uri;
    if (proxyinfo)
      this._proxyinfo = proxyinfo;

    // process item type and set contentType
    var sel = decodeURI(input_uri.path);
    if (sel == null || !sel.length || sel == '/') {
      this._itemtype = '1';
      this._selector = '';
    } else if (sel.length == 1) {
      this._itemtype = sel;
      this._selector = '';
    } else {
      this._itemtype = sel.substr(1,1);
      this._selector = sel.substr(2);
    }

    if (!input_uri.host || !input_uri.host.length)
      throw Cr.NS_ERROR_MALFORMED_URI;
    this._host = input_uri.host;
    if (!input_uri.port || input_uri.port < 1)
      this._port = 70;
    else
      this._port = input_uri.port;

    // force our itemtype. realistically, the old Gopher let people
    // slide a lot with content sniffing, but that really violated
    // the spirit of item types, so this is written to be as rigid as
    // possible for a particular itemtype (but see types I and s, sigh).
    var c;
    switch(this._itemtype) {
      case '1' :
      case '7' :
        c = GOPHER_MENU_MIME_TYPE;
        break;

      case '0' :
      case '2' :
        c = 'text/plain';
        break;

      case 'g' :
        c = 'image/gif' ;
        break;

      case 'I' :
        // oh, man, this is gross -- designed to
        // support both the common use of I for
        // JPEG images, and the official spec that
        // I is a 'general image type'
        //
        // taken from netwerk/mime/public/nsMimeTypes.h
        if (this._selector.match(/\.jpe?g$/i))
          c = 'image/jpeg' ;
        else if (this._selector.match(/\.gif$/i))
          c = 'image/gif' ; // grrRRR! use 'g'!!
        else if (this._selector.match(/\.xbm$/i))
          c = 'image/x-xbitmap' ;
        else if (this._selector.match(/\.png$/i))
          c = 'image/png' ;
        else if (this._selector.match(/\.svg$/i))
          c = 'image/svg+xml' ;
        else if (this._selector.match(/\.bmp$/i))
          c = 'image/bmp' ;
        else if (this._selector.match(/\.icon?$/i))
          c = 'image/x-icon' ;
        else if (this._selector.match(/\.tiff?$/i))
          c = 'image/tiff' ;
        else
          c = 'image/jpeg' ;
          // using application/octet-stream as the fallback broke
          // too many sites unfortunately
        break;

      case 'h' :
        c = 'text/html';
        break;

      case 'p' :
        c = 'image/png';
        break;

      case 'd' :
        c = 'application/pdf';
        break;

      case '8' :
      case 'T' :
        // shouldn't ever get this far (see GopherProtocol)
        throw Cr.NS_ERROR_NOT_IMPLEMENTED;
        break;

      default : // 9, 4, 5, 6, s, ;, etc.
        c = 'application/octet-stream';
        break;
    }
    this.contentType = c;
        
    GopherLog(("channel initialized: "+
      this._host + " " + this._port + " " +
      this._itemtype + " " +
      this._selector + " " +
      ""));
    return Cr.NS_OK;
  },
      
  open : function GC_open() {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },

  asyncOpen : function GC_asyncOpen(listener, context) {
    GopherLog(("trying to initialize transport"));

    var transportService =
                    Cc["@mozilla.org/network/socket-transport-service;1"].
                    getService(Ci.nsISocketTransportService);
    this._transport = transportService
      .createTransport(null, 0, this._host, this._port, this._proxyinfo);

    if (!(this.loadFlags & nsIRequest.LOAD_BACKGROUND)) {
      // hook up our event sink to the current UI thread
      var cq = Cc["@mozilla.org/thread-manager;1"].
               getService().currentThread;
      GopherLog("yes, we have sink "+cq);
      this._transport.setEventSink(this, cq);
    }

    // open and initialize the data pump to read from the socket
    var sinput = this._transport.openInputStream(0,0,0);
    this.sendRequest();
    this._pump = Cc["@mozilla.org/network/input-stream-pump;1"].
                 createInstance(Ci.nsIInputStreamPump);
    this._pump.init(sinput, -1, -1, 0, 0, true);
    this._pump.asyncRead(this, null);
    if (this._loadGroup) {
      this._loadGroup.addRequest(this, null);
      GopherLog("load group added");
    }

    this.isPending = true;
    this._listener = listener;
    this._context = context;

    // push on another content listener (in this case us) for
    // those itemtypes requiring translation to something else
    var transitives = [ '1', '7' ]; // item types for translation
    if (transitives.indexOf(this._itemtype) > -1) {
      GopherLog(("this type requires translation"));
      var dirconv = new GopherIndex();
      dirconv.asyncConvertData(GOPHER_MENU_MIME_TYPE,
                               'application/http-index-format',
                               this._listener, this._context);
      this._listener = dirconv;
      this._context = null;
      GopherLog(("now with dirconv: "+dirconv));
    }  

    GopherLog(("transport service for "+ this._host + " initialized"));
    return Cr.NS_OK;
  },

  get notificationCallbacks() {
    return this._notificationCallbacks;
  },
  set notificationCallbacks(nc) {
    this._notificationCallbacks = nc;
    this._progsink = null;
  },
  get securityInfo() {
    if (this._transport)
      return this._transport.securityInfo;
    return null;
  },
  // set securityInfo? in Gopher? I doubt it :)
  set securityInfo(foo) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    return null;
  },
  
  onStartRequest : function GC_onStartRequest(request, context) {
    if (this._listener)
      this._listener.onStartRequest(this, this._context);
    GopherLog(("onStartRequest "+this._listener));
  },

  onStopRequest : function GC_onStopRequest(request, context, status) {
    GopherLog(("onStopRequest: "+status));
    if(Components.isSuccessCode(status))
      this.status = status;

    if (this._listener) {
      this._listener.onStopRequest(this, this._context, this.status);
      GopherLog("listener stopped");
      this._listener = null;
      this._context = null;
    }
    if (this._loadGroup) {
      this._loadGroup.removeRequest(this, context, this.status);
      GopherLog("load group stopped");
    }

    this._pump = null;
    this._transport.close(this.status);
    this._transport = null;
    this.notificationCallbacks = null; // use our own getter/setter
    this._progsink = null;

    GopherLog("end of request");
    return Cr.NS_OK;
  },

  onDataAvailable : function GC_onDataAvailable(request, context, inputStream,
                                                offset, count) {
    GopherLog(("data event"));
    if (this._listener)
      this._listener.onDataAvailable(this, this._context,
                                     inputStream, offset, count);
    GopherLog(("data available: "+count+" bytes"));
  },
  sendRequest : function GC_sendRequest() {
    var transtring = this._selector;

    // the original version put up the itemtype 7 dialogue
    // at the channel, but I like bailing out early,
    // so we're doing that at the protocol handler level.
    // ditto for CSO/ph searches.

    if (this.csoargs && this.csoargs.length) {
      // completely replace selector with CSO/ph query
      // this eliminates a lot of headaches!!
      if (!this.csoargs.match(/^query /i)) // already query?
        transtring="query "+this.csoargs+" return all";
      else
        transtring=this.csoargs;
    } else if (this.queryargs && this.queryargs.length)
      transtring += "\t" + this.queryargs;
    transtring += "\r\n";
    // add terminating quit command to our query just in case
    // if this is CSO/ph
    if (this._itemtype == "2")
      transtring += "quit\r\n";

    // send the data
    var outstream = this._transport.openOutputStream(0,0,0);
    outstream.write(transtring, transtring.length);
    this.transreq = transtring; // for debugging
    outstream.close();
    GopherLog("selector sent: "+escape(transtring));
  },

  onTransportStatus : function GC_onTransportStatus(trans, status, prog,
                                                    progmax) {
    this._progsink = this._queryCallback(this, this._loadGroup,
                                         this._progsink);
    if (!this._progsink)
      GopherLog(("crap: no progsink"));
    GopherLog(("status changed: "+status+this._loadGroup+
      this.notificationCallbacks+this._progsink));

    if (this._progsink && !(this.loadFlags & nsIRequest.LOAD_BACKGROUND)){
      this._progsink.onStatus(this, this._context, status, this.URI.asciiHost);
      GopherLog(("onStatus"));
      
      if (status == nsISocketTransport.STATUS_RECEIVING_FROM ||
          status == nsISocketTransport.STATUS_SENDING_TO) {
        this._progsink.onProgress(this, this._context, prog, -1);
        GopherLog(("onProgress"));
      }
    }
    return Cr.NS_OK;
  }

};
 
/* The converter object for turning a gopher directory into an index.
   See http://www.mozilla.org/projects/netlib/dirindexformat.html .
   This is not separately exposed since GopherChannel is probably the only
   object that will ever call it.

   Currently, GopherIndex relies on the "magic" in
   netwerk/streamconv/converters/nsIndexedToHTML.cpp to format correctly. */

function GopherIndex() {
}

GopherIndex.prototype = {

  // my stuff
  _listener : null,
  _context : null,
  _menubuffer : '',
  _sstream : null,

  // internal functions
  _feedStream : function GIN_feedStream(request, listener, context, what) {
    var stringstream = Cc["@mozilla.org/io/string-input-stream;1"].
                       createInstance(Ci.nsIStringInputStream);
    stringstream.setData(what, what.length);
    listener.onDataAvailable(request, context, stringstream, 0, what.length);
  },

  // nsISupports
  QueryInterface: XPCOMUtils.generateQI([nsISupports,
                                         nsIRequestObserver,
                                         nsIStreamListener,
                                         nsIStreamConverter]),

  // nsIRequestObserver
  onStartRequest : function GIN_onStartRequest(request, context) {
    var myurl = (request && request.name && request.name.asciiSpec) ?
	request.name.asciiSpec : '';

    // init the channel with data and content type
    var chan = request.QueryInterface(nsIChannel);
    if (chan)
      chan.contentType = 'application/http-index-format';
    this._context = context;
    this._listener.onStartRequest(request, context);
    this._feedStream(request, this._listener, this._context,
	"300: "+myurl+ "\r\n" +
        "200: description filename file-type\r\n");
  },

  onStopRequest : function GIN_onStopRequest(request, context, status) {
    this._listener.onStopRequest(request, this._context, status);
    this._listener = null;
    this._context = null;
    if (this._sstream)
      this._sstream.close();
    this._sstream = null;
  },

  // nsIStreamListener
  onDataAvailable : function GIN_onDataAvailable(request, context, stream,
                                                 offset, count) {
    var i;
    var output = '';

    if (!this._sstream) {
      // create (and cache) our scriptable input stream
      this._sstream = Cc["@mozilla.org/scriptableinputstream;1"]
                      .createInstance(Ci.nsIScriptableInputStream);
      this._sstream.init(stream);
    }
    this._menubuffer += this._sstream.read(count);

    // pull the next newline-delimited string off the buffer
    while((i = this._menubuffer.indexOf("\n")) > -1) {
      var w = this._menubuffer.substr(0, i);
      if (i < this._menubuffer.length)
        this._menubuffer = this._menubuffer.substr(i+1);
      else
        this._menubuffer = '';

      w = w.replace("\r", "");
      w = w.replace("\n", "");

      // parse and break down the string into item type and tab-del'mted fields
      var itype = w.substr(0,1);
      var attribs = w.substr(1).split("\t");

      // display string
      var ds = attribs.shift();
      if (!ds)
        ds = ' ';
      ds = escape(ds);

      // selector
      var sel = attribs.shift(); // don't encode me yet
      if (!sel)
        sel = '';

      // hostname
      var host = attribs.shift();
      if (host && host.length)
        host = encodeURI(host);

      // port (maybe)
      var port = parseInt(attribs.shift());
      if (isNaN(port))
        port = 0; // falls through to bogosity filter

      // handle each item type

      if (itype == 'i' || itype == '3') { // informational lines
        output += "101: "+ds+"\r\n";

      } else if (host && host.length && port > 1) { // "true" item types
        var suburl;

        if (itype == '8' || itype == 'T') { // Telnet and TN3270
          suburl = (itype == '8') ? 'telnet' : 'tn3270';
          suburl += "://" + host + ":" + port + "/" + encodeURI(sel);
          output += "201: " + ds + " " + suburl + " FILE\r\n";

        } else // match hURL: syntax for URLs
          if (itype == 'h' &&
              (sel.substr(0,4) == "URL:" || sel.substr(0,5) == "/URL:")) {
          var subn = (sel.substr(0,1) == "/") ? 5 : 4;
          suburl = sel.substr(subn);
          // oh, master, don't let them feed us something bad!
          if (suburl.match(/^javascript:/) || suburl.match(/^data:/)) {
            output += "100: unsafe url "+suburl+"\r\n";
          } else {
            output += "201: " + ds + " " + encodeURI(suburl) + " FILE\r\n";
          }

        } else // match stealth HTTP requests and turn them into URLs
          if (port >= 80 && itype == 'h' && sel.match(/^[A-Z]+(%20| )/)) {
          sel = sel.replace(/^[A-Z]+(%20| )/,'');
          suburl = "http://" + host + ":" +
                   ((port != 80) ? (':' + port) : '') +
                   (sel.substr(0,1) == "/" ?  "" : "/") +
                   // the unescape is to deal with servers that
                   // have pre-encoded their URLs (typically most that
                   // still use this notation)
                   encodeURI(unescape(sel));
          output += "201: " + ds + " " + suburl + " FILE\r\n";

        } else { // for everything else, generate a regular URL
          suburl = GOPHER_PROTOCOL + '://' + host +
                   ((port != 70) ? (':' + port) : '') +
                   '/' + encodeURI(itype + sel);
          if (itype == "1" || itype == "7") {
            output += "201: " + ds + " " + suburl + " DIRECTORY\r\n";
          } else { 
            output += "201: " + ds + " " + suburl + " FILE\r\n";
          }
        }
      } else { // bogosity filter trap, log it silently for later
        output += "100: bogus element suppressed\r\n";
      }
    }
    this._feedStream(request, this._listener, this._context, output);
  },

  // nsIStreamConverter
  convert : function GIN_convert(from, to, listener, context) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },
  asyncConvertData : function GIN_asyncConvertData(from, to, listener,
                                                   context) {
    // if we were a real listener, we would do something with from and to,
    // but this will be left for future expansion perhaps.
    this._listener = listener;
    this._context = context;
  }
  
};

/* Finally, make our factory and module. */

var components = [ GopherProtocol ];
function NSGetModule(compMgr, fileSpec) {
  // why wouldn't ([ GopherProtocol ]) work, I wonder?
  return XPCOMUtils.generateModule(components);
}
